-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--                                                                         --
--                 MODULE Debug.hs    -     December 2002                  --
--                                                                         --
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

module Debug where

import Utilities
import Terms

--
--
-- Abstraction of terms
--
--

absSem k trs int = absSem' k trs [] int

absSem' _ _ _ [] = []
absSem' k trs prevInt int = intRef++absSem' k trs (prevInt++intRef) int'
   where intRef = (removeEqsFrom prevInt int)
         int' = (absStep k trs intRef)

absShow k trs int = absShow' k trs [] int

absShow' _ _ _ [] = []
absShow' k trs prevInt int = 
   intRef++((C "---" :-> C "---"):(absShow' k trs (prevInt++intRef) int'))
   where intRef = (removeEqsFrom prevInt int)
         int' = (absStep k trs intRef)
         
absSemIO k trs [] = putStrLn ""
absSemIO k trs int = 
   do showEquations int
      absSemIO k trs int'
   where int' = removeEqsFrom int (absStep k trs int)

{-
absSteps 0 k trs int = [int]
absSteps n k trs int = int:map (removeEqsFrom int) (absSteps (n-1) k trs int')
   where int' = (absStep k trs int)
-}
         
absStep k trs [] = []
absStep k trs (eq:int) = removeDuplicates ((absStepRules k eq trs) ++ (absStep k trs int))

absStepRules k (u :-> t) (sig,rules) = concat (map (absStepRulePos k u t [] sig) rules)

absStepRulePos k u t pos sig@(_,_,_,repFuncty) rule@(l :-> r) =
   if k<=length pos then [] else
   case subtermU of
      W x -> []
      V x -> case computedmgu of
                Nothing -> []
                Just sigma -> [(cutAt k (replaceSubTerm u pos (applySubstEqToTerm sigma l))) :-> t]
      C c -> if sortOfRule /= sortOfSubtermU then [] else
             case computedmgu of
                Nothing -> []
                Just sigma -> [(cutAt k (replaceSubTerm u pos (applySubstEqToTerm sigma l))) :-> t]
      F f ts -> if sortOfRule /= sortOfSubtermU then 
                   concat [absStepRulePos k u t (pos++[i]) sig (renamedRule (pos++[i]) rule) | i<-[1..length ts]] else 
                case computedmgu of
                    Nothing -> concat [absStepRulePos k u t (pos++[i]) sig (renamedRule (pos++[i]) rule) | i<-[1..length ts]]  
                    Just sigma -> ((cutAt k (applySubstEqToTerm sigma (replaceSubTerm u pos l))) :-> t):
                                  concat [absStepRulePos k u t (pos++[i]) sig (renamedRule (pos++[i]) rule) | i<-[1..length ts]]
   where
      functy = toFunc repFuncty
      sortOfRule = head (functy (rootOfTerm l))
      subtermU = subTerm u pos
      sortOfSubtermU = head (functy (rootOfTerm subtermU))
      computedmgu = mgu subtermU r
      renamedRule pos rule = renRuleWith ('#':show (foldr (+) 0 pos)) rule


--
--
--  Debugging
--
--

makeIdEqSpec set = [t :-> t | t<-set]


depthOfTerm (F f ts) = 1+ foldr1 max (map depthOfTerm ts)
depthOfTerm _ = 0

depthOfRule (l :-> r) = max (depthOfTerm l) (depthOfTerm r)

getK rules = foldr1 max (map depthOfRule rules)

testUnifyEqs (t:->s) (t':->s') = (testUnify t t') && (testUnify s s')

--
--
-- Finding incorrect rules
--
--

incorrectRule k iPlus iMinus sig rule = or [and [not (testUnifyEqs e e') |e'<-iPlus] | e<-oneStepOnRule]
   where oneStepOnRule = absStep k (sig,[renRuleWith "@" rule]) iMinus

incorrectRules blockingTerms trs@(sig,rules) spec@(_,eqs) = 
  [rule | rule<-rules, incorrectRule k overSpec underSpec sig rule]
   where 
      k = max (getK rules) (getK eqs)
      idBlockingTerms = makeIdEqSpec (map (cutAt k) blockingTerms)
      overSpec = absSem k spec idBlockingTerms
      underSpec = takeGroundEqs overSpec

incorrectRulesEval trs@(_,rules) spec@(_,eqs) = incorrectRules gCTerms trs spec
   where 
      k = max (getK rules) (getK eqs)
      gCTerms = makeGSTermsK (extractConstSignature spec) k

incorrectRulesGNF trs@(_,rules) spec@(_,eqs) = incorrectRules gNFs trs spec
   where 
      k = max (getK rules) (getK eqs)
      gNFs = makeGNFsK spec k

--
--
-- Finding uncovered equations
--
--

uncoveredEq e k iPlus (sig,rules) = and [not (testUnifyEqs e e') | e'<-oneStepWithRules]
   where oneStepWithRules = absStep k (sig,(map (renRuleWith "@") rules)) iPlus

uncoveredEquations blockingTerms trs@(_,rules) spec@(_,eqs) = 
   removeEqsFrom idBlockingTerms [e | e<-underSpec, uncoveredEq e k overSpec trs]
   where 
      k = max (getK rules) (getK eqs)
      idBlockingTerms = makeIdEqSpec (map (cutAt k) blockingTerms)
      overSpec = absSem k spec idBlockingTerms
      underSpec = takeGroundEqs overSpec

uncoveredEquationsEval trs@(_,rules) spec@(_,eqs) = uncoveredEquations gCTerms trs spec
   where 
      k = max (getK rules) (getK eqs)
      gCTerms = makeGSTermsK (extractConstSignature spec) k

uncoveredEquationsGNF trs@(_,rules) spec@(_,eqs) = uncoveredEquations gNFs trs spec
   where 
      k = max (getK rules) (getK eqs)
      gNFs = makeGNFsK spec k

--
--
-- Combined debugging
--
--

debugEval trs@(sig,rules) spec@(_,eqs) = (incorrectRules,uncoveredEqs)
   where 
      k = max (getK rules) (getK eqs)
      gCTermsK = map (cutAt k) (makeGSTermsK (extractConstSignature spec) k)
      idGCTerms = makeIdEqSpec (map (cutAt k) gCTermsK)
      overSpec = absSem k spec idGCTerms
      underSpec = takeGroundEqs overSpec
      incorrectRules = [rule | rule<-rules, incorrectRule k overSpec underSpec sig rule]
      uncoveredEqs = removeEqsFrom idGCTerms [e | e<-underSpec, uncoveredEq e k overSpec trs]
    
debugGNF trs@(sig,rules) spec@(_,eqs) = (incorrectRules,uncoveredEqs)
   where 
      k = max (getK rules) (getK eqs)
      gNFsK = map (cutAt k) (makeGNFsK spec k)
      idgNFs = makeIdEqSpec (map (cutAt k) gNFsK)
      overSpec = absSem k spec idgNFs
      underSpec = takeGroundEqs overSpec
      incorrectRules = [rule | rule<-rules, incorrectRule k overSpec underSpec sig rule]
      uncoveredEqs = removeEqsFrom idgNFs [e | e<-underSpec, uncoveredEq e k overSpec trs]
