-- To run the tool execute "main"
module Main where

import On_demand_rewriting
import OBJ_parser
import OBJ_write
import Time
import IO
import Parsec(parse)

-- Main function
main = 
  do{ putStrLn "           \\||||||||||||||||||||||||||||||||||||||||/"
    ; putStrLn "     --- Welcome to On-demand prototype for OBJ Family ---"
    ; putStrLn "           /||||||||||||||||||||||||||||||||||||||||\\"
    ; putStrLn "                 Version 1 built: Nov 18 2002"
    ; putStrLn "                 Copyright 2002 U.P. Valencia"
    ; today <- getClockTime
    ; todayCalendar <- toCalendarTime today
    ; putStrLn ("                 "++calendarTimeToString todayCalendar)
    ; putStrLn ""
    ; putStrLn "Commands: in filename --Load file and select module"
    ; putStrLn "          red expr.   --Reduce expression"
    ; putStrLn "          mred expr.  --Reduce with multiple normal forms"
    ; putStrLn "          tred expr.  --Traced reduction"
    ; putStrLn "          norm expr.  --Normalize expression via mu-normalization"
    ; putStrLn "          list        --Print loaded program"
    ; putStrLn "          quit        --Quit"
    ; mainsh (Obj "init" [] [] [] [])
  }

mainsh :: OBJProgram -> IO ()
mainsh objp = 
  do{ putStr ("\n"++idOBJP objp++"> ")
    ; hFlush stdout
    ; line <- getLine
    ; menu objp line
  }

menu :: OBJProgram -> String -> IO ()
menu objp1 ('i':'n':' ':filename) = 
  do{ objp <- parseOBJ filename
    ; if ((idOBJP $! objp)/="error")
      then do{ putStrLn ("Module "++idOBJP objp++" loaded")
             ; mainsh objp
           }
      else do{ putStrLn "Error parsing"
             ; mainsh objp1
           }
  }

menu objp ('r':'e':'d':' ':sexpr) = 
  do{ expr <- parseExpr sexpr
    ; if (expr==Var "") then putStr ""
      else ondemandEvalOpt objp (label objp expr)
    ; mainsh objp
  }

menu objp ('m':'r':'e':'d':' ':sexpr) =
  do{ expr <- parseExpr sexpr
    ; if (expr==Var "") then putStr ""
      else ondemandEval False objp [] [(label objp expr,[],0)]
    ; mainsh objp
  }

menu objp ('t':'r':'e':'d':' ':sexpr) = 
  do{ expr <- parseExpr sexpr
    ; if (expr==Var "") then putStr ""
      else ondemandEval True objp [] [(label objp expr,[],0)]
    ; mainsh objp
  }

menu objp ('n':'o':'r':'m':' ':sexpr) = 
  do{ expr <- parseExpr sexpr
    ; if (expr==Var "") then putStr ""
      else normalize objp (label objp expr)
    ; mainsh objp
  }

menu _ ('q':'u':'i':'t':[]) = putStrLn "Bye."

menu objp "list" = putStrLn (strOBJ objp) >> mainsh objp

menu objp _ = 
  do{ putStrLn "Error: command not understood"
    ; mainsh objp
  }

idOBJP (Obj id _ _ _ _) = id

parseExpr :: String -> IO TermD
parseExpr sexpr = 
  do{ case (parse term "prompt" (init sexpr)) of
           Left err -> do{ putStr "Parse error in expression at "
                         ; print err
                         ; return (Var "")
                         }
           Right x  -> return x
  }

