-----------------------------------------------------------------------------

-----------------------------------------------------------------------------

--                                                                         --

--              MODULE debussy.hs  v. 1.0 - December 2002                  --

--                                                                         --

-----------------------------------------------------------------------------

-----------------------------------------------------------------------------



import Utilities

import Terms

import OBJ_parser

import Debug



--------------------------------------------------------------

--

-- From Parsec/OBJ representation to local representation

--

--------------------------------------------------------------



termsToTerm (Var x) = V x

termsToTerm (Call f []) = C f

termsToTerm (Call f terms) = F f (map termsToTerm terms)



opsToFuncs (Op f _ _ _) = f

varsToVars (x,_) = x

opsToArity (Op f _ sorts _) = (f,(length sorts)-1)



--

-- opsToFuncty :: OpD -> ([Char],[[Char]])

-- 

-- Obtains the functionality of an operator. The first sort

-- in the stored list is the return sort

--



opsToFuncty (Op f _ sorts _) = (f,(last sorts):init sorts)

eqsToRules (Eq t s) = (termsToTerm t) :-> (termsToTerm s)



objToTRS (Obj _ sorts ops vars eqs) = (signature,rules)

   where 

    signature = (map opsToFuncs ops,map varsToVars vars,

                 map opsToArity ops,

                 map opsToFuncty ops)

    rules = map eqsToRules eqs

           



    

--------------------------------------------------------------

--

--  User interface

--

--------------------------------------------------------------





askForString msg noConfirm existing =

   do putStrLn msg

      s <- getLine

      if s == "" then (return Nothing) else 

         if elem s existing then 

            do putStrLn "Repeated labels are not allowed!"

               askForString msg noConfirm existing

         else

            if (elem s noConfirm) then return (Just s) else

               do putStrLn ("Please, confirm the input '"++s++"' (y/n) :")

                  cs <- getLine

                  if (toLower (head cs))=='y' then return (Just s) else 

                     askForString msg noConfirm existing





header = ["\n","\n",

   "      D E B U G G I N G   M E E T S   R E W R I T I N G   S Y S T E M S\n",

   "-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --\n",

   "\n",

   "              ||                                   ||\n",

   "              ||                 _      _          ||\n",

   "              ||__________     //-\\\\   /_\\___\\     ||----------\n",

   "              ||                   \\\\_// \\---/     ||----------\n",

   "              ||                     -             ||\n",

   "              ||                                   ||\n",  

   "\n",              

   "              DEBUSSY            Version 1.0       December 2002\n",

   "-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --\n",

   "      D E B U g g i n g   m e e t S   r e w r i t i n g   S Y s t e m s",

   "\n","\n"]

type Menu = [String]

type Label = String

type LabelledDebProblem id = ((Label,Maybe [RewRule id]),(Label,Maybe [Equation id]))

type LabelledTRSs id = (Label,TRS id)

type Store id = ([LabelledDebProblem id],[LabelledTRSs id])

data SelMenu = MMain | MFile | MLabels | MView | MDebug | MDebugEval | MDebugGNF



selMenu MMain = (menuMain,processMenuMain)

selMenu MFile = (menuFile,processMenuFile)

selMenu MLabels = (menuLabels,processMenuLabels)

selMenu MView = (menuView,processMenuView)

selMenu MDebug = (menuDebug,processMenuDebug)

selMenu MDebugEval = (menuDebugEval,processMenuDebugEval)

selMenu MDebugGNF = (menuDebugGNF,processMenuDebugGNF)



menuMain = 

   [">>>   M A I N   <<<",

    "",

    "1. File...",

    "2. Labels...",

    "3. View...",

    "",

    "4. Debug...",

    "",

    "0. Exit"]



processMenuMain '1' store = return (False,MFile,store)

processMenuMain '2' store@(_,[]) = 

      do putStrLn "No current TRS"

         return (False,MMain,store) 

processMenuMain '2' store = return (False,MLabels,store) 

processMenuMain '3' store@(_,[]) = 

      do putStrLn "No current TRS"

         return (False,MMain,store) 

processMenuMain '3' store = return (False,MView,store) 

processMenuMain '4' store@(_,[]) = 

      do putStrLn "No current TRS"

         return (False,MMain,store) 

processMenuMain '4' store = return (False,MDebug,store) 

processMenuMain '0' store = 

   do quitting <- askForString "Do you want to leave the current session? (y/n)" ["y","n","Y","N"] []

      case quitting of

         Just "y" -> return (True,MMain,store)

         Just "Y" -> return (True,MMain,store)

         otherwise -> return (False,MMain,store)

processMenuMain _ store = return (False,MMain,store)



menuFile = 

   [">>>   F I L E   <<<",

    "",

    "1. Load TRS in simple format",

    "2. Load OBJ specification in simple format",

    "",

    "9. Set current TRS",

    "",

    "0. Back"]



processMenuFile '0' store = return (False,MMain,store)

processMenuFile '1' store@(debug,eqs) = 

   do name <- askForString "Name of the file ('Enter' quits): " [] []

      if name == Nothing then return (False,MFile,store) else 

         let Just fileName = name in

             do fileTRS <- readFile fileName

                label <- askForString "Label of the TRS ('Enter' uses default): " [] []

                let labelTRS = case label of

                                Nothing -> "trs"

                                Just inputLabel -> inputLabel in

                    return (False,MFile,(debug,(labelTRS,obtainTRS fileTRS):eqs))

processMenuFile '2' store@(debug,eqs) = 

   do name <- askForString "Name of the file ('Enter' quits): " [] []

      if name == Nothing then return (False,MFile,store) else 

         let Just fileName = name in

             do label <- askForString "Label of the OBJ specification ('Enter' uses default): " [] []

                objSpec <- parseOBJ fileName

                let labelEQs = case label of

                                Nothing -> "obj"

                                Just inputLabel -> inputLabel in

                    return (False,MFile,(debug,(labelEQs,objToTRS objSpec):eqs))

processMenuFile '9' store@(_,[_]) = 

      do putStrLn "There is only one TRS"

         return (False,MFile,store) 

processMenuFile '9' store@(debug,eqs0@((label,_):eqs)) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++(foldr (++) [] (numberedLabels (map fst eqs))))

      eqs' <- setNewCurrentTRS eqs0

      return (False,MFile,(debug,eqs'))            

processMenuFile _ store = return (False,MFile,store)



menuLabels = 

   [">>>   L A B E L S   <<<",

    "",

    "1. Show labels (column)",

    "2. Show labels (row)",

    "",

    "3. Remove TRS",

    "4. Remove all but current TRS",

    "",

    "8. Clear store",

    "",    

    "9. Set current TRS",

    "",

    "0. Back"]



removePair label = filter ((label/=).fst)



rmvItem i [] = []

rmvItem 0 (x:xs) = xs

rmvItem i (x:xs) = x:rmvItem (i-1) xs



numberedLabels ids = 

   [' ':' ':' ':(show i)++('.':' ':label) | (i,label) <- zip [1..] ids]



setNewCurrentItem msg items = 

   do putStr msg

      ns <- getLine

      if (ns == "" || any (not.isDigit) ns) then (return items) else  

         let i = strToNum ns 

         in if i==0 || i>(length items) then return items else 

               return ((items !! i):rmvItem i items)

               

setNewCurrentTRS trss = 

   setNewCurrentItem "New current TRS? ('n' selects; 'Enter' quits) " trss

   

removeItem msg items = 

   do putStr msg

      ns <- getLine

      if (ns == "" || any (not.isDigit) ns) then (return items) else  

         let i = strToNum ns

         in if i==0 then (return items) else 

               return ((head items):(rmvItem i (tail items)))

               

processMenuLabels '0' store = return (False,MMain,store)

processMenuLabels _ store@(_,[]) = 

   do putStrLn "No TRS loaded"

      return (False,MLabels,store)         

processMenuLabels '1' store@(_,(label,_):eqs) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++(foldr (++) [] (map (("\n    "++).fst) eqs)))

      return (False,MLabels,store)

processMenuLabels '2' store@(_,(label,_):eqs) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++(foldr (++) [] (map (("    "++).fst) eqs)))

      return (False,MLabels,store)

processMenuLabels '3' store@(_,[_]) = 

   do putStrLn "The current TRS cannot be removed"

      return (False,MLabels,store) 

processMenuLabels '3' (debug,eqs0@((label,_):eqs)) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++

               (foldr (++) [] (numberedLabels (map fst eqs))))

      eqs' <- removeItem msg eqs0

      return (False,MLabels,(debug,eqs'))

   where

      msg = "Delete TRS? ('n' selects; 'Enter' quits) "    

processMenuLabels '4' store@(_,[_]) = return (False,MLabels,store)

processMenuLabels '4' store@(debug,currentTRS:_) = 

   do removing <- askForString "Remove all but current TRS? (y/n)" ["y","n","Y","N"] []

      case removing of

         Just "y" -> return (False,MLabels,(debug,[currentTRS]))

         Just "Y" -> return (False,MLabels,(debug,[currentTRS]))

         otherwise -> return (False,MLabels,store)            

processMenuLabels '8' store@(debug,eqs) = 

   do removing <- askForString "Clear store? (y/n)" ["y","n","Y","N"] []

      case removing of

         Just "y" -> return (False,MLabels,([],[]))

         Just "Y" -> return (False,MLabels,([],[]))

         otherwise -> return (False,MLabels,store)            

processMenuLabels '9' store@(_,[_]) = 

      do putStrLn "There is only one TRS"

         return (False,MLabels,store) 

processMenuLabels '9' store@(debug,eqs0@((label,_):eqs)) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++(foldr1 (++) (numberedLabels (map fst eqs))))

      eqs' <- setNewCurrentTRS eqs0

      return (False,MLabels,(debug,eqs'))            

processMenuLabels _ store = return (False,MLabels,store)





menuView = 

   [">>>   V I E W   <<<",

    "",

    "1. Signature",

    "2. Rules",

    "",

    "3. TRS",

    "",

    "9. Set current TRS",

    "",

    "0. Back"]



processMenuView '0' store = return (False,MMain,store)

processMenuView _ store@(_,[]) = 

   do putStrLn "No TRS loaded"

      return (False,MMain,store)         

processMenuView '1' store@(_,(_,trs):_) = 

   do showSignature sig

      return (False,MView,store)

   where (sig,_) = trs

processMenuView '2' store@(_,(_,trs):_) = 

   do showRules rules

      return (False,MView,store)

   where (_,rules) = trs

processMenuView '3' store@(_,(label,trs):_) = 

   do putStrLn ('\n':"Showing the TRS  "++label++":") 

      showTRS trs

      return (False,MView,store)

processMenuView '9' store@(_,[_]) = 

      do putStrLn "There is only one TRS"

         return (False,MView,store) 

processMenuView '9' store@(debug,eqs0@((label,_):eqs)) = 

   do putStrLn ("Current TRS: "++label++"\nOthers:   "++(foldr (++) [] (numberedLabels (map fst eqs))))

      eqs' <- setNewCurrentTRS eqs0

      return (False,MView,(debug,eqs'))            

processMenuView _ store = return (False,MView,store)



menuDebug = 

   [">>>   D E B U G   <<<",

    "",

    "1. Use the evaluation semantics...",

    "2. Use the (ground) normalization semantics...",

    "",

    "4. New debugging problem",

    "5. Show debugging problems",

    "6. Reset current debugging problem",

    "7. Remove debugging problem",

    "",

    "8. Clear store",

    "",    

    "9. Set current debugging problem",

    "",

    "0. Back"]



newDebugProblem debProblems trss =

   do putStrLn "Defining a new debugging problem: "

      spec <- newSpecification trss

      prog <- newProgram trss

      if ((spec == Nothing) || (prog == Nothing)) then (return debProblems) else 

         let Just trs' = prog

             Just spec' = spec 

         in return ((trs',spec'):debProblems)

newSpecification trss = 

   do putStrLn ("Please, select the specification ('n' selects; 'Enter' quits) "++numLabels)

      ns <- getLine

      if ((ns == "") || (any (not.isDigit) ns)) then (return Nothing) else 

         let i = strToNum ns 

         in if i==0 || i>(length trss) then (return Nothing) else 

            return (Just (fst (trss !! (i-1)),Nothing))

   where numLabels = foldr (++) [] (numberedLabels (map fst trss))

newProgram trss = 

   do putStrLn ("Please, select the program ('n' selects; 'Enter' quits) "++numLabels)

      ns <- getLine

      if ((ns == "") || (any (not.isDigit) ns)) then (return Nothing)  else 

         let i = strToNum ns

         in if i==0 || i>(length trss) then (return Nothing) else 

            return (Just (fst (trss !! (i-1)),Nothing))

   where numLabels = foldr (++) [] (numberedLabels (map fst trss))

   

showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs =

   do putStrLn ("Incorrect rules in " ++ labelProg ++ " :")

      showRules incorrectRules

      putStrLn ("Uncovered equations from " ++ labelSpec ++ " :")

      showEquations uncoveredEqs

      

setNewCurrentDebugProblem debProblems = 

   setNewCurrentItem "New current debugging problem? ('n' selects; 'Enter' quits) " debProblems

   

showDebuggingProblems (((labelProg,incRules),(labelSpec,uncEqs)):debProblems) =

   do putStrLn "Current debugging problem is:\n"

      putStrLn ("   Specification: "++labelSpec++statusDeb uncEqs)

      putStrLn ("   Program      : "++labelProg++statusDeb incRules)

      putStrLn ""; putStrLn "Others:\n"

      putStrLn (foldr (++) [] (numberedLabels (map extractLabels debProblems)))



statusDeb Nothing = " (?)"

statusDeb (Just _) = " (!)"



extractLabels ((l,incRules),(l',uncEqs)) = 

   '(':l'++(statusDeb uncEqs)++',':l++(statusDeb incRules)++")\n"



processMenuDebug :: Char -> Store String -> IO (Bool,SelMenu,Store String)

processMenuDebug '0' store = return (False,MMain,store)

processMenuDebug _ store@(_,[]) = 

   do putStrLn "No TRS loaded"

      return (False,MMain,store)

processMenuDebug '1' store = return (False,MDebugEval,store)

processMenuDebug '2' store = return (False,MDebugGNF,store)

processMenuDebug '4' store@(_,[_]) = 

   do putStrLn "At least two TRSs are necessary!"

      return (False,MDebug,store) 

processMenuDebug '4' store@(debProblems,trss) = 

   do debProblems' <- newDebugProblem debProblems trss

      return (False,MDebug,(debProblems',trss))            

processMenuDebug '5' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebug,store)

processMenuDebug '5' store@(debProblems,_) = 

   do showDebuggingProblems debProblems

      return (False,MDebug,store)      

processMenuDebug '6' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebug,store)

processMenuDebug '6' store@(((labelProg,_),(labelSpec,_)):debProblems,trss) = 

   do resetting <- askForString "Reset current debugging problem? (y/n)" ["y","n","Y","N"] []

      case resetting of

         Just "y" -> return newStore

         Just "Y" -> return newStore

         otherwise -> return (False,MDebug,store)

   where 

      newStore = (False,MDebug,(((labelProg,Nothing),(labelSpec,Nothing)):debProblems,trss))    

processMenuDebug '7' store@([_],_) = 

      do putStrLn "The current debugging problem cannot be removed"

         return (False,MDebug,store) 

processMenuDebug '7' store@(debProblems,trss) = 

   do showDebuggingProblems debProblems

      debProblems' <- removeItem msg debProblems

      return (False,MDebug,(debProblems',trss))

   where 

      msg = "Delete debugging problem? ('n' selects; 'Enter' quits) "

processMenuDebug '8' store@(debug,eqs) = 

   do removing <- askForString "Clear store? (y/n)" ["y","n","Y","N"] []

      case removing of

         Just "y" -> return (False,MDebug,([],[]))

         Just "Y" -> return (False,MDebug,([],[]))

         otherwise -> return (False,MDebug,store)            

processMenuDebug '9' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebug,store)

processMenuDebug '9' store@([_],_) = 

   do putStrLn "There is only one debugging problem"

      return (False,MDebug,store)

processMenuDebug '9' (debProblems,trss) = 

   do showDebuggingProblems debProblems

      debProblems' <- setNewCurrentDebugProblem debProblems

      return (False,MDebug,(debProblems',trss))

processMenuDebug _ store = return (False,MDebug,store)



menuDebugEval = 

   [">>>   D E B U G  - Evaluation Semantics  <<<",

    "",

    "1. Debugging w.r.t. the evaluation semantics",

    "2.    Show incorrect rules only",

    "3.    Show uncovered equations only",

    "",

    "9. Set current debugging problem",

    "",

    "0. Back"]



processMenuDebugEval '0' store = return (False,MDebug,store)

processMenuDebugEval _ store@(_,[]) = 

   do putStrLn "No TRS loaded"

      return (False,MMain,store)

processMenuDebugEval '1' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugEval,store)

processMenuDebugEval '1' (((labelProg,Nothing),(labelSpec,Nothing)):debProblems,trss) = 

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugEval,store)

   where

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      (incorrectRules,uncoveredEqs) = debugEval prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugEval '1' (((labelProg,Just incorrectRules),(labelSpec,Nothing)):debProblems,trss) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugEval,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      uncoveredEqs = uncoveredEquationsEval prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugEval '1' (((labelProg,Nothing),(labelSpec,Just uncoveredEqs)):debProblems,trss) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugEval,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      incorrectRules = incorrectRulesEval prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugEval '1' store@(((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):_,_) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugEval,store)

processMenuDebugEval '2' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugEval,store)

processMenuDebugEval '2' (((labelProg,Nothing),uncovEqs@(labelSpec,_)):debProblems,trss) = 

   do putStrLn ("Incorrect rules in " ++ labelProg ++ " :")

      showRules incorrectRules

      return (False,MDebugEval,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      incorrectRules = incorrectRulesEval prog eqs

      store = (((labelProg,Just incorrectRules),uncovEqs):debProblems,trss)

processMenuDebugEval '2' store@(((labelProg,Just incorrectRules),_):_,_) = 

   do putStrLn ("Incorrect rules in " ++ labelProg ++ " :")

      showRules incorrectRules

      return (False,MDebugEval,store)

processMenuDebugEval '3' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugEval,store)

processMenuDebugEval '3' ((incRules@(labelProg,_),(labelSpec,Nothing)):debProblems,trss) = 

   do putStrLn ("Uncovered equations from " ++ labelSpec ++ " :")

      showEquations uncoveredEqs

      return (False,MDebugEval,store)

   where

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      uncoveredEqs = uncoveredEquationsEval prog eqs

      store = ((incRules,(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugEval '3' store@((_,(labelSpec,Just uncoveredEqs)):_,_) = 

   do putStrLn ("Uncovered equations from " ++ labelSpec ++ " :")

      showEquations uncoveredEqs

      return (False,MDebugEval,store)

processMenuDebugEval '9' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugEval,store)

processMenuDebugEval '9' store@([_],_) = 

   do putStrLn "There is only one debugging problem"

      return (False,MDebugEval,store)

processMenuDebugEval '9' (debProblems,trss) = 

   do showDebuggingProblems debProblems

      debProblems' <- setNewCurrentDebugProblem debProblems

      return (False,MDebugEval,(debProblems',trss))

processMenuDebugEval _ store = return (False,MDebugEval,store)





menuDebugGNF = 

   [">>>   D E B U G  - Ground Normal Forms Semantics  <<<",

    "",

    "1. Debugging w.r.t. the ground normal forms semantics",

    "2.    Show incorrect rules only",

    "3.    Show uncovered equations only",

    "",

    "9. Set current debugging problem",

    "",

    "0. Back"]



processMenuDebugGNF '0' store = return (False,MDebug,store)

processMenuDebugGNF _ store@(_,[]) = 

   do putStrLn "No TRS loaded"

      return (False,MMain,store)

processMenuDebugGNF '1' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugGNF,store)

processMenuDebugGNF '1' (((labelProg,Nothing),(labelSpec,Nothing)):debProblems,trss) = 

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugGNF,store)

   where

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      (incorrectRules,uncoveredEqs) = debugGNF prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugGNF '1' (((labelProg,Just incorrectRules),(labelSpec,Nothing)):debProblems,trss) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugGNF,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      uncoveredEqs = uncoveredEquationsGNF prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugGNF '1' (((labelProg,Nothing),(labelSpec,Just uncoveredEqs)):debProblems,trss) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugGNF,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      incorrectRules = incorrectRulesGNF prog eqs

      store = (((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugGNF '1' store@(((labelProg,Just incorrectRules),(labelSpec,Just uncoveredEqs)):_,_) =

   do showDebuggingResults labelProg labelSpec incorrectRules uncoveredEqs

      return (False,MDebugGNF,store)

processMenuDebugGNF '2' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugGNF,store)

processMenuDebugGNF '2' (((labelProg,Nothing),uncovEqs@(labelSpec,_)):debProblems,trss) = 

   do putStrLn ("Incorrect rules in " ++ labelProg ++ " :")

      showRules incorrectRules

      return (False,MDebugGNF,store)

   where 

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      incorrectRules = incorrectRulesGNF prog eqs

      store = (((labelProg,Just incorrectRules),uncovEqs):debProblems,trss)

processMenuDebugGNF '2' store@(((labelProg,Just incorrectRules),_):_,_) = 

   do putStrLn ("Incorrect rules in " ++ labelProg ++ " :")

      showRules incorrectRules

      return (False,MDebugGNF,store)

processMenuDebugGNF '3' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugGNF,store)

processMenuDebugGNF '3' ((incRules@(labelProg,_),(labelSpec,Nothing)):debProblems,trss) = 

   do putStrLn ("Uncovered equations from " ++ labelSpec ++ " :")

      showEquations uncoveredEqs

      return (False,MDebugGNF,store)

   where

      Just prog = lookup labelProg trss

      Just eqs = lookup labelSpec trss

      uncoveredEqs = uncoveredEquationsGNF prog eqs

      store = ((incRules,(labelSpec,Just uncoveredEqs)):debProblems,trss)

processMenuDebugGNF '3' store@((_,(labelSpec,Just uncoveredEqs)):_,_) = 

   do putStrLn ("Uncovered equations from " ++ labelSpec ++ " :")

      showEquations uncoveredEqs

      return (False,MDebugGNF,store)

processMenuDebugGNF '9' store@([],_) = 

   do putStrLn "No current debugging problem"

      return (False,MDebugGNF,store)

processMenuDebugGNF '9' store@([_],_) = 

   do putStrLn "There is only one debugging problem"

      return (False,MDebugGNF,store)

processMenuDebugGNF '9' (debProblems,trss) = 

   do showDebuggingProblems debProblems

      debProblems' <- setNewCurrentDebugProblem debProblems

      return (False,MDebugGNF,(debProblems',trss))

processMenuDebugGNF _ store = return (False,MDebugGNF,store)



--

-- 

-- Menu management

--

--



initialStore = ([],[])



viewMenu menu = do putStrLn "\n\n"

                   putStrLn (foldr (++) [] (map ('\n':) menu))

                   putChar '\n'



--

-- Shortcuts for menu options

--



accessToMenu '0' = MMain

accessToMenu '1' = MFile

accessToMenu '2' = MLabels

accessToMenu '3' = MView



accessToMenu 'f' = MFile

accessToMenu 'l' = MLabels

accessToMenu 'v' = MView

accessToMenu 'd' = MDebug

accessToMenu 'e' = MDebugEval

accessToMenu 'g' = MDebugGNF



accessToMenu _ = MMain



continue selM store =

   let (menu,fmenu) = selMenu selM

   in do 

        viewMenu menu

        cs <- getLine

        case cs of

             "" -> (continue selM store) 

             'm':n:_ -> if isDigit n then continue (accessToMenu n) store 

                        else (continue selM store)

             'M':n:_ -> if isDigit n then continue (accessToMenu n) store

                        else (continue selM store)

             menu:_ -> if isAlpha menu then continue (accessToMenu (toLower menu)) store

                       else do (end,selM',store') <- fmenu (head cs) store

                               if end then (putStrLn "\n\nBye!") else (continue selM' store')

debussy = 

   do putStrLn (foldr1 (++) header) 

      continue MMain initialStore

      putStr (foldr1 (++) header)

      putStrLn "Please, report any bug, comment, or suggestion to: slucas@dsic.upv.es" 



